
#include <_sio_internal.h>

#define	_SIO_EXTREP_VALMASK	(INT_MAX >> _SIO_EXTREP_USIZE)
#define	_SIO_EXTREP_VAL(ext)	((ext) & _SIO_EXTREP_VALMASK)
#define	_SIO_EXTREP_CREATE(u,v)	(((u) << _SIO_EXTREP_USHIFT) | (v))

struct _sio_exttable {
	pthread_mutex_t mutex;
	TAILQ_HEAD(, _sio_exttable_entry) queue;
	unsigned int uniq, next;
};

struct _sio_exttable_entry {
	TAILQ_ENTRY(_sio_exttable_entry) link;
	int external;
	void *internal;
};

static struct _sio_exttable_entry *
    _sio_exttable_find_internal(struct _sio_exttable *, void *internal);
static struct _sio_exttable_entry *
    _sio_exttable_find_external(struct _sio_exttable *, int external);

void *
_sio_exttable_create(unsigned int tableuniq, unsigned int expected)
{
	struct _sio_exttable *net;

	net = malloc(sizeof *net);
	if (net == NULL)
		return (NULL);

	pthread_mutex_init(&net->mutex, pthread_mutexattr_default);
	TAILQ_INIT(&net->queue);
	net->uniq = tableuniq;
	net->next = 0;

	/* XXX expected would be used if hash table, etc. */

	return (net);
}

static struct _sio_exttable_entry *
_sio_exttable_find_internal(struct _sio_exttable *tp, void *internal)
{
	struct _sio_exttable_entry *ep;

	for (ep = tp->queue.tqh_first; ep != NULL; ep = ep->link.tqe_next)
		if (ep->internal == internal)
			return (ep);
	return NULL;
}

static struct _sio_exttable_entry *
_sio_exttable_find_external(struct _sio_exttable *tp, int external)
{
	struct _sio_exttable_entry *ep;

	for (ep = tp->queue.tqh_first; ep != NULL; ep = ep->link.tqe_next)
		if (ep->external == external)
			return (ep);
	return NULL;
}

void
_sio_exttable_register(void *tpv, void *internal)
{
	struct _sio_exttable *tp = tpv;
	struct _sio_exttable_entry *ep;
	int external;

	if (internal == NULL) {
		_sio_panic("_sio_exttable_register: NULL");
		/* NOTREACHED */
	}

	ep = malloc(sizeof *ep);
	if (ep == NULL)
		_sio_panic("_sio_exttable_register: out of memory");

	ep->internal = internal;

	pthread_mutex_lock(&tp->mutex);

	do {
		/* try the next external representation */
		external = _SIO_EXTREP_CREATE(tp->uniq, tp->next);

		/* increment the value */
		tp->next++;
		tp->next &= _SIO_EXTREP_VALMASK;
	} while (_sio_exttable_find_external(tp, external));

	ep->external = external;

	_sio_rcobj_ref(internal);
	TAILQ_INSERT_TAIL(&tp->queue, ep, link);

	pthread_mutex_unlock(&tp->mutex);
}

void
_sio_exttable_unregister(void *tpv, void *internal)
{
	struct _sio_exttable *tp = tpv;
	struct _sio_exttable_entry *ep;

	if (internal == NULL) {
		_sio_panic("_sio_exttable_destroy: NULL");
		/* NOTREACHED */
	}

	pthread_mutex_lock(&tp->mutex);

	ep = _sio_exttable_find_internal(tp, internal);
	if (ep == NULL)
		_sio_panic("_sio_exttable_unregister: %p not in table %p",
		    internal, tp);
	TAILQ_REMOVE(&tp->queue, ep, link);
	_sio_rcobj_unref(internal);

	pthread_mutex_unlock(&tp->mutex);

	free(ep);
}

int
_sio_exttable_externalize(void *tpv, void *internal)
{
	struct _sio_exttable *tp = tpv;
	struct _sio_exttable_entry *ep;
	int rv;

	if (internal == NULL) {
		_sio_panic("_sio_exttable_externalize: NULL");
		/* NOTREACHED */
	}

	pthread_mutex_lock(&tp->mutex);

	ep = _sio_exttable_find_internal(tp, internal);
	if (ep == NULL)
		_sio_panic("_sio_exttable_externalize: %p not in table %p",
		    internal, tp);

	rv = ep->external;
	_sio_rcobj_unref(internal);

	pthread_mutex_unlock(&tp->mutex);
	return (rv);
}

void *
_sio_exttable_internalize(void *tpv, int external)
{
	struct _sio_exttable *tp = tpv;
	struct _sio_exttable_entry *ep;
	void *rv;

	rv = NULL;
	pthread_mutex_lock(&tp->mutex);

	ep = _sio_exttable_find_external(tp, external);
	if (ep != NULL) {
		_sio_rcobj_ref(ep->internal);
		rv = ep->internal;
	} else {
		rv = NULL;
	}

	pthread_mutex_unlock(&tp->mutex);
	return (rv);
}
